@@ -7,11 +7,13 @@ import (
 	"sort"
 	"strings"
 
-	"github.com/StackExchange/dnscontrol/v3/models"
-	"github.com/StackExchange/dnscontrol/v3/pkg/diff2"
-	"github.com/StackExchange/dnscontrol/v3/pkg/js"
-	"github.com/StackExchange/dnscontrol/v3/pkg/printer"
+	"github.com/StackExchange/dnscontrol/v4/models"
+	"github.com/StackExchange/dnscontrol/v4/pkg/diff2"
+	"github.com/StackExchange/dnscontrol/v4/pkg/js"
+	"github.com/StackExchange/dnscontrol/v4/pkg/printer"
 	"github.com/urfave/cli/v2"
+
+	"github.com/fatih/color"
 )
 
 // categories of commands
@@ -22,7 +24,11 @@ const (
 )
 
 var commands = []*cli.Command{}
-var version string
+
+// These are set by/for goreleaser
+var (
+	version = "dev"
+)
 
 func cmd(cat string, c *cli.Command) bool {
 	c.Category = cat
@@ -62,6 +68,18 @@ func Run(v string) int {
 			Name:        "diff2",
 			Usage:       "Enable replacement diff algorithm",
 			Destination: &diff2.EnableDiff2,
+			Value:       true,
+		},
+		&cli.BoolFlag{
+			Name:        "disableordering",
+			Usage:       "Disables the dns ordering part of the diff2 package",
+			Destination: &diff2.DisableOrdering,
+		},
+		&cli.BoolFlag{
+			Name:        "no-colors",
+			Usage:       "Disable colors",
+			Destination: &color.NoColor,
+			Value:       false,
 		},
 	}
 	sort.Sort(cli.CommandsByName(commands))
