@@ -5,13 +5,14 @@ import (
 	"encoding/json"
 	"fmt"
 	"io"
+	"log"
 	"net/http"
 	"os"
 	"sort"
 	"strings"
 	"time"
 
-	"github.com/StackExchange/dnscontrol/v3/pkg/printer"
+	"github.com/StackExchange/dnscontrol/v4/pkg/printer"
 	"github.com/mattn/go-isatty"
 )
 
@@ -19,8 +20,6 @@ const apiBase = "https://apis.cscglobal.com/dbs/api/v2"
 
 // Api layer for CSC Global
 
-type requestParams map[string]string
-
 type errorResponse struct {
 	Code        string `json:"code"`
 	Description string `json:"description"`
@@ -644,6 +643,12 @@ func (client *providerClient) geturl(url string) ([]byte, error) {
 	req.Header.Add("Authorization", "Bearer "+client.token)
 	req.Header.Add("Accept", "application/json")
 
+	// Default CSCGlobal rate limit is twenty requests per second
+	var backoff = time.Second
+
+	const maxBackoff = time.Second * 15
+
+retry:
 	resp, err := hclient.Do(req)
 	if err != nil {
 		return nil, err
@@ -656,6 +661,22 @@ func (client *providerClient) geturl(url string) ([]byte, error) {
 
 	if resp.StatusCode == 400 {
 		// 400, error message is in the body as plain text
+		// Apparently CSCGlobal uses status code 400 for rate limit, grump
+
+		if string(bodyString) == "Requests exceeded API Rate limit." {
+			// a simple exponential back-off with a 3-minute max.
+			log.Printf("Delaying %v due to ratelimit\n", backoff)
+			time.Sleep(backoff)
+			backoff = backoff + (backoff / 2)
+			if backoff > maxBackoff {
+				return nil, fmt.Errorf("CSC Global API timeout max backoff (geturl): %s URL: %s%s",
+					bodyString,
+					req.Host, req.URL.RequestURI())
+			}
+
+			goto retry
+		}
+
 		return nil, fmt.Errorf("CSC Global API error (geturl): %s URL: %s%s",
 			bodyString,
 			req.Host, req.URL.RequestURI())
