@@ -3,7 +3,7 @@ package vultr
 import (
 	"testing"
 
-	"github.com/StackExchange/dnscontrol/v3/models"
+	"github.com/StackExchange/dnscontrol/v4/models"
 	"github.com/vultr/govultr/v2"
 )
 
@@ -72,7 +72,3 @@ func TestConversion(t *testing.T) {
 		}
 	}
 }
-
-func intPtr(v int) *int {
-	return &v
-}
